# Generated by Django 2.2.4 on 2021-11-12 12:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('function_tools', '0002_registeredfunction_tags'),
    ]

    operations = [
        migrations.CreateModel(
            name='ImplementationStrategy',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
            ],
            options={
                'verbose_name': 'Стратегия создания функции',
                'db_table': 'function_tools_implementation_strategy',
            },
        ),
    ]



# Generated by Django 2.2.4 on 2021-10-26 23:37

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('function_tools', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='registeredfunction',
            name='tags',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=128), default=list, size=None),
        ),
    ]



from typing import (
    Generic,
    Iterator,
    TypeVar,
)

from django.db.models import (
    QuerySet,
)

_Z = TypeVar("_Z")


class QueryType(Generic[_Z], QuerySet):
    """
    Используется при типизировании QuerySet для того, чтобы интерпритатор
    понимал, с каким объектами нужно работать при итерировании

    Пример использования:
    acc_analytics: QueryType[AccountAnalytic] = AccountAnalytic.objects.filter(...
    """

    def __iter__(self) -> Iterator[_Z]: ...



from typing import (
    Type,
    Union,
)

from function_tools.helpers import (
    BaseFunctionHelper,
    BaseHelper,
    BaseRunnerHelper,
)
from function_tools.validators import (
    BaseValidator,
)


class HelperMixin:
    """
    Миксин для появления помощника у сущности

    В каждый класс, использующий данный миксин должна быть добавлена

    self._helper: Union[BaseHelper, BaseFunctionHelper] = self._prepare_helper(*args, **kwargs)

    в __init__(*args, **kwargs)
    """

    @property
    def helper(self):
        """
        Возвращает помощника
        """
        return self._helper

    def _prepare_helper_class(self) -> Union[Type[BaseHelper], Type[BaseFunctionHelper]]:
        """
        Возвращает класс помощника
        """
        return BaseHelper

    def _prepare_helper(self, *args, **kwargs):
        """
        Точка расширения для создания помощника.
        """
        helper_class = self._prepare_helper_class()

        if issubclass(helper_class, (BaseHelper, BaseFunctionHelper)):
            helper = helper_class(*args, **kwargs)
        else:
            helper = BaseHelper(*args, **kwargs)

        return helper


class ValidatorMixin:
    """
    Миксин для появления валидатора у сущности

    У класса, использующего миксин необходимо добавить

    self._validator: BaseValidator = self._prepare_validator(*args, **kwargs)

    в __init__(*args, **kwargs)
    """

    def _prepare_validator_class(self) -> Type[BaseValidator]:
        """
        Возвращает класс валидатор
        """
        return BaseValidator

    def _prepare_validator(self, *args, **kwargs):
        """
        Точка расширения для создания валидатора
        """
        validator_class = self._prepare_validator_class()

        if issubclass(validator_class, BaseValidator):
            validator = validator_class(*args, **kwargs)
        else:
            validator = BaseValidator(*args, **kwargs)

        return validator

    def before_validate(self):
        """
        Расширение поведения запускаемого объекта перед запуском проверок
        """

    def validate(self):
        """
        Публичный метод для запуска валидатора сущности
        """
        if self._validator:
            self._validator.validate(self)

    def after_validate(self):
        """
        Возможность расширения запускаемого объекта после валидации
        """


class GlobalHelperMixin:
    """
    Миксин добавляющий возможность установки глобального помощника

    У классов наследников должна быть добавлена

    self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)

    в __init__(*args, **kwargs)
    """

    @property
    def global_helper(self):
        """
        Возвращает глобального помощника
        """
        return self._global_helper

    def _prepare_global_helper_class(self) -> Type[BaseRunnerHelper]:
        """
        Возвращает класс помощника
        """
        return BaseRunnerHelper

    def _prepare_global_helper(self, *args, **kwargs):
        """
        Предназначен для определения глобального помощника
        """
        global_helper_class = self._prepare_global_helper_class()

        if issubclass(global_helper_class, BaseRunnerHelper):
            global_helper = global_helper_class(*args, **kwargs)
        else:
            global_helper = BaseRunnerHelper()

        return global_helper

    def set_global_helper(
        self,
        global_helper: BaseRunnerHelper,
    ):
        """
        Публичный метод для установки глобального помощника
        """
        self._global_helper = global_helper



from typing import (
    Optional,
    Type,
    Union,
)

from function_tools.caches import (
    BaseCache,
    CacheStorage,
)


class BaseHelper:
    """
    Базовый класс помощника. Предполагается, что у наследников будут
    создаваться кеши и собираться прочая вспомогательная информация, которая
    требуется для исполнения действий
    """

    def __init__(self, *args, **kwargs):
        super().__init__()

        self._cache_class = self._prepare_cache_class()

        self._cache: Union[Optional[BaseCache], Optional[CacheStorage]] = None

        self._prepare_cache(*args, **kwargs)

    @property
    def cache(self):
        """
        Возвращает кеш помощника
        """
        return self._cache

    def _prepare_cache_class(self) -> Union[Optional[Type[BaseCache]], Optional[Type[CacheStorage]]]:
        """
        Возвращает класс кеша помощника
        """
        return BaseCache

    def _prepare_cache(self, *args, **kwargs):
        """
        Метод создания кеша.

        Кеш хранится в публичном свойстве cache. По умолчанию добавлена
        заглушка.
        """
        if issubclass(self._cache_class, (BaseCache, CacheStorage)):
            self._cache = self._cache_class(*args, **kwargs)
        else:
            self._cache = BaseCache(*args, **kwargs)


class BaseRunnerHelper(BaseHelper):
    """
    Базовый класс помощника пусковика выполнения функций
    """


class BaseFunctionHelper(BaseHelper):
    """
    Базовый класс помощника функции. Предполагается, что в наследниках будут
    создаваться кеши и собираться прочая вспомогательная информация, которая
    требуется для исполнения функции
    """
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)



from functools import (
    wraps,
)


def run_without_errors(func):
    """
    Декоратор методов функции запускаемых только в случае отсутствия зарегистрированных ошибок в процессе работы
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.result.has_not_errors:
            return func(self, *args, **kwargs)

    return wrapper



import os
import shutil
import sys
from distutils.util import (
    strtobool,
)
from importlib import (
    import_module,
)
from os import (
    path,
)
from pathlib import (
    Path,
)

import django
from django.conf import (
    settings,
)
from django.core.management.base import (
    CommandError,
)
from django.core.management.templates import (
    TemplateCommand,
)
from django.core.management.utils import (
    handle_extensions,
)
from django.template import (
    Context,
    Engine,
)
from django.utils.version import (
    get_docs_version,
)
from isort.api import (
    sort_code_string,
)
from isort.settings import (
    DEFAULT_CONFIG,
)

import function_tools
from function_tools.management.consts import (
    PARAMETERS_DIALOG_WINDOW,
)
from function_tools.management.storages import (
    ImplementationStrategyStorage,
)
from function_tools.models import (
    ImplementationStrategy,
)


class PatchedTemplateCommand(TemplateCommand):
    """
    Пропатченная команда для создания пакетов по шаблону
    """

    rewrite_template_suffixes = (
        ('.py-tpl', '.py'),
        ('.js-tpl', '.js'),
        ('.md-tpl', '.md'),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.name = None
        self.app_or_project = None
        self.paths_to_remove = None
        self.verbosity = None
        self.base_name = None
        self.base_subdir = None
        self.base_directory = None
        self.base_python_path = None
        self.camel_case_name = None
        self.camel_case_value = None
        self.extra_files = None
        self.extensions = None

        self._top_dir_path = None
        self._python_path_value = None

    def _get_conf_dir_parent_path(self):
        """
        Возвращает абсолютный путь директории, содержащей директорию conf с шаблонами внутри
        """
        return django.__path__[0]

    def _sort_imports(
        self,
        content: str,
    ):
        """
        Сортировка импортов при помощи isort
        """
        return sort_code_string(
            code=content,
            config=getattr(settings, 'ISORT_CONFIG') or DEFAULT_CONFIG,
        )

    def handle_template(self, template=None, subdir=None):
        """
        Поиск директории с шаблоном функции
        """
        template_directory_path = None

        for app_name in settings.INSTALLED_APPS:
            app_module = import_module(app_name)
            app_path = app_module.__path__

            application_path = Path(app_path[0])

            temp_template_directory_path = (
                application_path / 'function_templates' / self.strategy.function_template_name
            )

            if temp_template_directory_path.is_dir():
                template_directory_path = temp_template_directory_path
                break

        if not template_directory_path:
            raise CommandError(f'couldn\'t handle {self.strategy.function_template_name} template.')

        return template_directory_path

    def _make_top_dir(self, target, name):
        """
        Создание корневой директории
        """
        # if some directory is given, make sure it's nicely expanded
        if target is None:
            self._top_dir_path = path.join(os.getcwd(), name)
            try:
                os.makedirs(self._top_dir_path)
            except FileExistsError:
                raise CommandError(f'\'{self._top_dir_path}\' already exists')
            except OSError as e:
                raise CommandError(e)
        else:
            self._top_dir_path = os.path.abspath(path.expanduser(target))
            if not os.path.exists(self._top_dir_path):
                raise CommandError(
                    f'Destination directory \'{self._top_dir_path}\' does not exist, please create it first.'
                )

    def _prepare_top_module_python_path(self):
        """
        Формирование пути создаваемого пакета для дальнейшего использования в генерируемых импортах
        """
        top_dir_path = self._top_dir_path
        paths = [
            path_item
            for path_item in sys.path
            if path_item in top_dir_path
            # при запуске через django-admin в Python path добавляется директория, из которой запускается команда, что
            # искажает полноценный путь
            and f'{path_item}/{self.name}' != top_dir_path
        ]

        max_sys_path = max(paths)

        self._python_path_value = top_dir_path.split(f'{max_sys_path}/')[1]

        if '/' in self._python_path_value:
            self._python_path_value = self._python_path_value.replace('/', '.')

    def _prepare_extra_files(self, options):
        self.extra_files = []
        for file in options['files']:
            self.extra_files.extend(map(lambda x: x.strip(), file.split(',')))

    def _prepare_extensions(self, options):
        """
        Подготовка расширений файлов
        """
        self.extensions = tuple(handle_extensions(options['extensions']))

        if self.verbosity >= 2:
            self.stdout.write(
                f'Rendering {self.app_or_project} template files with extensions: {", ".join(self.extensions)}\n'
            )

            self.stdout.write(
                f'Rendering {self.app_or_project} template files with filenames: {", ".join(self.extra_files)}\n'
            )

    def _prepare_new_path_file(self, filename, relative_dir, options):
        """
        Подготовка пути генерируемого из шаблона файла
        """
        new_path = path.join(self._top_dir_path, relative_dir, filename.replace(self.base_name, self.name))

        for old_suffix, new_suffix in self.rewrite_template_suffixes:
            if new_path.endswith(old_suffix):
                new_path = new_path[:-len(old_suffix)] + new_suffix
                break  # Only rewrite once

        if path.exists(new_path):
            raise CommandError(
                f"{new_path} already exists, overlaying a project or app into an existing directory won't "
                f"replace conflicting files"
            )

        return new_path

    def _render_file(self, new_path, old_path, filename):
        # Only render the Python files, as we don't want to
        # accidentally render Django templates files
        if new_path.endswith(self.extensions) or filename in self.extra_files:
            with open(old_path, 'r', encoding='utf-8') as template_file:
                content = template_file.read()
            template = Engine().from_string(content)
            content = template.render(self.context)
            with open(new_path, 'w', encoding='utf-8') as new_file:
                if new_path.endswith('.py'):
                    content = self._sort_imports(content)

                new_file.write(content)
        else:
            shutil.copyfile(old_path, new_path)

        if self.verbosity >= 2:
            self.stdout.write("Creating %s\n" % new_path)
        try:
            shutil.copymode(old_path, new_path)
            self.make_writeable(new_path)
        except OSError:
            self.stderr.write(
                msg=(
                    f'Notice: Couldn\'t set permission bits on {new_path}. You\'re probably using an uncommon '
                    f'filesystem setup. No problem.'
                ),
                style_func=self.style.NOTICE,
            )

    def _create_package_by_template(self, options):
        template_dir = str(self.handle_template())
        prefix_length = len(template_dir) + 1

        template_dir_items = [(root, dirs, files) for root, dirs, files in os.walk(template_dir)]

        if not template_dir_items:
            raise CommandError('Please, check template directory, because directory is empty!')

        for root, dirs, files in template_dir_items:

            path_rest = root[prefix_length:]
            relative_dir = path_rest.replace(self.base_name, self.name)
            if relative_dir:
                target_dir = path.join(self._top_dir_path, relative_dir)
                if not path.exists(target_dir):
                    os.mkdir(target_dir)

            for dirname in dirs[:]:
                if dirname.startswith('.') or dirname == '__pycache__':
                    dirs.remove(dirname)

            for filename in files:
                if filename.endswith(('.pyo', '.pyc', '.py.class')):
                    # Ignore some files as they cause various breakages.
                    continue
                old_path = path.join(root, filename)

                new_path = self._prepare_new_path_file(filename, relative_dir, options)

                self._render_file(new_path, old_path, filename)

    def _prepare_context(self, options):
        """
        Создание контекста
        """
        self.context = Context({
            **options,
            self.base_name: self.name,
            self.base_directory: self._top_dir_path,
            self.camel_case_name: self.camel_case_value,
            self.base_python_path: self._python_path_value,
            'docs_version': get_docs_version(),
            'django_version': django.__version__,
        }, autoescape=False)

    def _django_setup(self):
        """
        Инициализация Django для рендеринга шаблонов
        """
        if not settings.configured:
            settings.configure()
            django.setup()

    def _remove_paths(self):
        """
        Удаление помеченных файлов и директорий
        """
        if self.paths_to_remove:
            if self.verbosity >= 2:
                self.stdout.write('Cleaning up temporary files.\n')
            for path_to_remove in self.paths_to_remove:
                if path.isfile(path_to_remove):
                    os.remove(path_to_remove)
                else:
                    shutil.rmtree(path_to_remove)

    def _prepare_parameters(self, app_or_project, name, options):
        """
        Подготовка параметров для дальнейшей работы
        """
        self.name = name
        self.app_or_project = app_or_project
        self.paths_to_remove = []
        self.verbosity = options['verbosity']
        self.base_name = f'{app_or_project}_name'
        self.strategy = options['strategy']
        self.base_directory = f'{app_or_project}_directory'
        self.base_python_path = f'{app_or_project}_python_path'
        self.camel_case_name = f'camel_case_{app_or_project}_name'
        self.camel_case_value = ''.join(x for x in name.title() if x != '_')
        self.validate_name(name, app_or_project)

    def handle(self, app_or_project, name, target=None, **options):
        """
        Template command handler
        """
        self._prepare_parameters(app_or_project, name, options)
        self._make_top_dir(target=target, name=name)
        self._prepare_top_module_python_path()
        self._prepare_extra_files(options)
        self._prepare_extensions(options)
        self._prepare_context(options)
        self._create_package_by_template(options)
        self._remove_paths()


class Command(PatchedTemplateCommand):
    """
    Команда для создания Функции приложения
    """

    help = (
        """
        Команда для создания Функции приложения. В неименованном параметре передается наименование функции в стиле 
        camel_case. Переданное имя будет служить в качестве наименования пакета. Также будет произведено 
        преобразование переданного наименования к виду CamelCase, которое будет использоваться в наименовании классов.
        
        Существует несколько стратегий реализации Функции. Указать необходимую можно при помощи параметра --strategy.
        """
    )

    missing_args_message = "Вы должны указать наименование функции."

    def add_arguments(self, parser):
        """
        Добавление параметров команды
        """
        super().add_arguments(parser)

        strategy_help = '\n'.join([
            f'{strategy.key} - {strategy.title};'
            for strategy in ImplementationStrategy.get_enum_data().values()
        ])

        parser.add_argument(
            '--strategy',
            dest='strategy',
            action='store',
            type=lambda s: str(s).upper(),
            help=(
                f"""
                Стратегия реализации Функции.
                
                Может принимать следующие значения:
                {strategy_help} 
                """
            ),
            default=ImplementationStrategy.SYNC_LAZY_SAVING_FUNCTION.key,
        )

        parser.add_argument(
            '--verbose_name',
            dest='function_verbose_name',
            action='store',
            type=str,
            help='Полное (человекочитаемое) название функции',
            default='Безымянная функция',
        )

        parser.add_argument(
            '--is_parameterized',
            dest='is_parameterized',
            action='store',
            type=strtobool,
            help='Является ли функция параметризированной с необходимостью показа окна с параметрами пользователю?',
            default=False,
        )

    def _get_conf_dir_parent_path(self):
        """
        Возвращает абсолютный путь директории, содержащей директорию conf с шаблонами внутри
        """
        return function_tools.__path__[0]

    def _make_top_dir(self, target, name):
        super()._make_top_dir(target, name)

        if 'functions' not in self._top_dir_path:
            raise RuntimeError('Path for creating function must contain "functions" directory')

    def _prepare_new_path_file(self, filename, relative_dir, options):
        """
        Подготовка пути генерируемого из шаблона файла
        """
        new_path = super()._prepare_new_path_file(filename, relative_dir, options)

        if PARAMETERS_DIALOG_WINDOW in new_path:
            if options['is_parameterized']:
                new_path = new_path.replace(PARAMETERS_DIALOG_WINDOW, f'{self.camel_case_value}{PARAMETERS_DIALOG_WINDOW}')  # noqa
            else:
                new_path = None

        return new_path

    def _render_file(self, new_path, old_path, filename):
        if new_path:
            super()._render_file(new_path, old_path, filename)

    def _prepare_base_subdir_parameter(self, app_or_project, options):
        """
        Формирование параметра base_subdir
        """
        function_type = options['function_type']

        self.base_subdir = f'{app_or_project}_{function_type.name.lower()}_template'

    def _prepare_parameters(self, app_or_project, name, options):
        """
        Подготовка параметров для дальнейшей работы
        """
        super()._prepare_parameters(app_or_project, name, options)

        self.url_name = self.name.replace('_', '-')
        self.base_url_name = f'{app_or_project}_url_name'

    def _prepare_context(self, options):
        """
        Создание контекста
        """
        self.context = Context({
            **options,
            self.base_name: self.name,
            self.base_url_name: self.url_name,
            self.base_directory: self._top_dir_path,
            self.camel_case_name: self.camel_case_value,
            self.base_python_path: self._python_path_value,
            'docs_version': get_docs_version(),
            'django_version': django.__version__,
        }, autoescape=False)

    def handle(self, **options):
        function_name = options.pop('name')
        target = options.pop('directory')

        strategy_key = options.pop('strategy')

        strategy_storage = ImplementationStrategyStorage()
        strategy = strategy_storage.get_strategy_implementation(
            strategy_key=strategy_key,
        )

        options['strategy'] = strategy

        options['extensions'].extend([
            'md',
            'js',
        ])

        super().handle('function', function_name, target, **options)



from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Optional,
)

from function_tools.caches import (
    CacheStorage,
)
from function_tools.errors import (
    BaseError,
)
from function_tools.functions import (
    BaseFunction,
    LazyDelegateSavingPredefinedQueueFunction,
    LazySavingPredefinedQueueFunction,
)
from function_tools.helpers import (
    BaseFunctionHelper,
    BaseRunnerHelper,
)
from function_tools.managers import (
    RunnerManager,
)
from function_tools.presenters import (
    ResultPresenter,
)
from function_tools.results import (
    BaseRunnableResult,
)
from function_tools.runners import (
    BaseRunner,
    LazySavingRunner,
)
from function_tools.validators import (
    BaseValidator,
)


class FunctionImplementationStrategy(ABC):
    """
    Базовый класс стратегии реализации функции
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        self._manager_class = None
        self._runner_class = None
        self._function_class = None
        self._runner_helper_class = None
        self._function_helper_class = None
        self._runner_validator_class = None
        self._function_validator_class = None
        self._runner_cache_storage_class = None
        self._function_cache_storage_class = None
        self._error_class = None
        self._runner_result_class = None
        self._function_result_class = None
        self._result_presenter_class = None

        self._key = self._prepare_key()
        self._title = self._prepare_title()
        self._function_template_name = self._prepare_function_template_name()

        self._prepare()

    @property
    def key(self) -> Optional[str]:
        """
        Возвращает уникальный идентификатор стратегии создания функции
        """
        return self._key

    @property
    def title(self) -> Optional[str]:
        """
        Возвращает название стратегии создания функции
        """
        return self._title

    @property
    def function_template_name(self) -> Optional[str]:
        """
        Возвращает наименование стратегии реализации функции
        """
        return self._function_template_name

    @property
    def manager_class(self):
        return self._manager_class

    @property
    def manager_class_name(self):
        return self._manager_class.__name__

    @property
    def manager_class_module(self):
        return self._manager_class.__module__

    @property
    def runner_class(self):
        return self._runner_class

    @property
    def runner_class_name(self):
        return self._runner_class.__name__

    @property
    def runner_class_module(self):
        return self._runner_class.__module__

    @property
    def function_class(self):
        return self._function_class

    @property
    def function_class_name(self):
        return self._function_class.__name__

    @property
    def function_class_module(self):
        return self._function_class.__module__

    @property
    def runner_helper_class(self):
        return self._runner_helper_class

    @property
    def runner_helper_class_name(self):
        return self._runner_helper_class.__name__

    @property
    def runner_helper_class_module(self):
        return self._runner_helper_class.__module__

    @property
    def function_helper_class(self):
        return self._function_helper_class

    @property
    def function_helper_class_name(self):
        return self._function_helper_class.__name__

    @property
    def function_helper_class_module(self):
        return self._function_helper_class.__module__

    @property
    def runner_validator_class(self):
        return self._runner_validator_class

    @property
    def runner_validator_class_name(self):
        return self._runner_validator_class.__name__

    @property
    def runner_validator_class_module(self):
        return self._runner_validator_class.__module__

    @property
    def function_validator_class(self):
        return self._function_validator_class

    @property
    def function_validator_class_name(self):
        return self._function_validator_class.__name__

    @property
    def function_validator_class_module(self):
        return self._function_validator_class.__module__

    @property
    def runner_cache_storage_class(self):
        return self._runner_cache_storage_class

    @property
    def runner_cache_storage_class_name(self):
        return self._runner_cache_storage_class.__name__

    @property
    def runner_cache_storage_class_module(self):
        return self._runner_cache_storage_class.__module__

    @property
    def function_cache_storage_class(self):
        return self._function_cache_storage_class

    @property
    def function_cache_storage_class_name(self):
        return self._function_cache_storage_class.__name__

    @property
    def function_cache_storage_class_module(self):
        return self._function_cache_storage_class.__module__

    @property
    def error_class(self):
        return self._error_class

    @property
    def error_class_name(self):
        return self._error_class.__name__

    @property
    def error_class_module(self):
        return self._error_class.__module__

    @property
    def runner_result_class(self):
        return self._runner_result_class

    @property
    def runner_result_class_name(self):
        return self._runner_result_class.__name__

    @property
    def runner_result_class_module(self):
        return self._runner_result_class.__module__

    @property
    def function_result_class(self):
        return self._function_result_class

    @property
    def function_result_class_name(self):
        return self._function_result_class.__name__

    @property
    def function_result_class_module(self):
        return self._function_result_class.__module__

    @property
    def result_presenter_class(self):
        return self._result_presenter_class

    @property
    def result_presenter_class_name(self):
        return self._result_presenter_class.__name__

    @property
    def result_presenter_class_module(self):
        return self._result_presenter_class.__module__

    @abstractmethod
    def _prepare_key(self) -> Optional[str]:
        """
        Формирование уникального ключа стратегии создания функции
        """

    @abstractmethod
    def _prepare_title(self) -> Optional[str]:
        """
        Формирование наименования стратегии создания функции
        """
    @abstractmethod
    def _prepare_function_template_name(self) -> Optional[str]:
        """
        Формирование названия шаблона создания функции
        """

    def _prepare_manager_class(self):
        """
        Устанавливает класс менеджера
        """
        self._manager_class = RunnerManager

    def _prepare_runner_class(self):
        """
        Устанавливает класс пускателя
        """
        self._runner_class = BaseRunner

    def _prepare_function_class(self):
        """
        Устанавливает класс Функции
        """
        self._function_class = BaseFunction

    def _prepare_runner_helper_class(self):
        """
        Устанавливает класс помощника пусковика
        """
        self._runner_helper_class = BaseRunnerHelper

    def _prepare_function_helper_class(self):
        """
        Устанавливает класс помощника функции
        """
        self._function_helper_class = BaseFunctionHelper

    def _prepare_runner_validator_class(self):
        """
        Устанавливает класс валидатора пусковика
        """
        self._runner_validator_class = BaseValidator

    def _prepare_function_validator_class(self):
        """
        Устанавливает класс валидатора Функции
        """
        self._function_validator_class = BaseValidator

    def _prepare_runner_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей пусковика
        """
        self._runner_cache_storage_class = CacheStorage

    def _prepare_function_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей Функции
        """
        self._function_cache_storage_class = CacheStorage

    def _prepare_error_class(self):
        """
        Устанавливает класс ошибки
        """
        self._error_class = BaseError

    def _prepare_runner_result_class(self):
        """
        Устанавливает класс результата
        """
        self._runner_result_class = BaseRunnableResult

    def _prepare_function_result_class(self):
        """
        Устанавливает класс результата
        """
        self._function_result_class = BaseRunnableResult

    def _prepare_result_presenter_class(self):
        """
        Устанавливает класс презентера результата
        """
        self._result_presenter_class = ResultPresenter

    def _prepare(self):
        """
        Подготовка компонентов реализации функции
        """
        self._prepare_manager_class()
        self._prepare_runner_class()
        self._prepare_function_class()
        self._prepare_runner_helper_class()
        self._prepare_function_helper_class()
        self._prepare_runner_validator_class()
        self._prepare_function_validator_class()
        self._prepare_function_cache_storage_class()
        self._prepare_runner_cache_storage_class()
        self._prepare_error_class()
        self._prepare_runner_result_class()
        self._prepare_function_result_class()
        self._prepare_result_presenter_class()


class SyncBaseRunnerBaseFunctionImplementationStrategy(FunctionImplementationStrategy):
    """
    Реализация простой функции без отложенного сохранения
    """

    def _prepare_key(self) -> str:
        """
        Возвращает уникальный идентификатор стратегии создания функции
        """
        return 'SYNC_BASE_FUNCTION'

    def _prepare_title(self) -> str:
        """
        Возвращает название стратегии создания функции
        """
        return 'Реализация простой функции без отложенного сохранения'

    def _prepare_function_template_name(self) -> Optional[str]:
        """
        Формирование названия шаблона создания функции
        """
        return 'function_sync_template'


class SyncBaseRunnerLazySavingPredefinedQueueFunctionImplementationStrategy(FunctionImplementationStrategy):
    """
    Реализация функции с отложенным сохранением и предустановленной очередью объектов на сохранение
    """

    def _prepare_key(self) -> str:
        """
        Возвращает уникальный идентификатор стратегии создания функции
        """
        return 'SYNC_LAZY_SAVING_FUNCTION'

    def _prepare_title(self) -> str:
        """
        Возвращает название стратегии создания функции
        """
        return (
            'Реализация функции с отложенным сохранением и предустановленной очередью объектов на сохранение. '
            'Сохранение производится после удачной работы функции'
        )

    def _prepare_function_template_name(self) -> Optional[str]:
        """
        Формирование названия шаблона создания функции
        """
        return 'function_sync_template'

    def _prepare_function_class(self):
        """
        Устанавливает класс Функции
        """
        self._function_class = LazySavingPredefinedQueueFunction


class SyncLazySavingRunnerLazyDelegateSavingPredefinedQueueFunctionImplementationStrategy(FunctionImplementationStrategy):
    """
    Реализация функции с отложенным сохранением его делегированием пускателю. Когда все функции отработают, только
    после этого запускается сохранение объектов из очередей каждой функции
    """

    def _prepare_key(self) -> str:
        """
        Возвращает уникальный идентификатор стратегии создания функции
        """
        return 'SYNC_LAZY_SAVING_RUNNER_FUNCTION'

    def _prepare_title(self) -> str:
        """
        Возвращает название стратегии создания функции
        """
        return (
            'Реализация функции с отложенным сохранением его делегированием пускателю. Когда все функции отработают, '
            'только после этого запускается сохранение объектов из очередей каждой функции'
        )

    def _prepare_function_template_name(self) -> Optional[str]:
        """
        Формирование названия шаблона создания функции
        """
        return 'function_sync_template'

    def _prepare_runner_class(self):
        """
        Устанавливает класс пусковика
        """
        self._runner_class = LazySavingRunner

    def _prepare_function_class(self):
        """
        Устанавливает класс Функции
        """
        self._function_class = LazyDelegateSavingPredefinedQueueFunction



from abc import (
    ABCMeta,
    abstractmethod,
)
from collections import (
    Iterable,
    Sequence,
    deque,
)
from typing import (
    Deque,
    List,
    Optional,
    Union,
)

from function_tools.consts import (
    EMPTY_VERBOSE_NAME,
)
from function_tools.general import (
    LazySavingActionModelRunnableObject,
    RunnableObject,
)
from function_tools.helpers import (
    BaseFunctionHelper,
    BaseHelper,
    BaseRunnerHelper,
)
from function_tools.mixins import (
    GlobalHelperMixin,
    HelperMixin,
)


class BaseFunction(
    HelperMixin,
    RunnableObject,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций
    """

    # Полное наименование функции для дальнейшей регистрации и отображения пользователю
    verbose_name: str = EMPTY_VERBOSE_NAME

    # Список тегов функции, по которым ее можно в дальнейшем отыскать
    tags: List[str] = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._helper: Union[BaseHelper, BaseFunctionHelper] = self._prepare_helper(*args, **kwargs)

    @abstractmethod
    def _prepare(self):
        """
        Метод содержащий все действия работы функции
        """

    def run(self):
        """
        Контрактный метод служащий для запуска функции
        """
        self._prepare()


class BaseGlobalHelperFunction(
    GlobalHelperMixin,
    BaseFunction,
    metaclass=ABCMeta,
):
    """Метод содержащий все действия работы функции с глобальным помощником."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)


class LazySavingFunction(
    BaseFunction,
    LazySavingActionModelRunnableObject,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания классов функций с отложенным сохранением
    """

    def __init__(self, *args, ignore_errors_on_saving: bool = False, **kwargs):
        """
        Инициализация функции с отложенным сохранением

        Args:
            ignore_errors_on_saving Указывает на игнорирование наличия ошибок при выполнении запроса и необходимость
            сохранения результата в БД в любом случае
        """
        self._ignore_errors_on_saving = ignore_errors_on_saving

        super().__init__(*args, **kwargs)

    def do_on_save(self, object_):
        """
        Добавление действия на момент сохранения

        Принимает действие (список, экземпляр модели, функция)
        """
        if isinstance(object_, (Iterable, Sequence)):
            for act in object_:
                self.do_on_save(act)
        else:
            self._queue_to_save.append(object_)

    def run(self):
        """
        Выполнение действий функции с дальнейшим сохранением объектов в базу
        при отсутствии ошибок или явном указании игнорирования ошибок при сохранении
        """
        self._prepare()

        if self._ignore_errors_on_saving or self.result.has_not_errors:
            self.do_save()


class LazySavingPredefinedQueueFunction(
    LazySavingFunction,
    metaclass=ABCMeta
):
    """
    Базовый класс для создания функций с отложенным сохранением объектов
    моделей с предустановленной очередью
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._queue: Optional[Deque] = deque()


class LazySavingPredefinedQueueGlobalHelperFunction(
    GlobalHelperMixin,
    LazySavingPredefinedQueueFunction,
    metaclass=ABCMeta
):
    """
    Базовый класс для создания функций с отложенным сохранением,
    предустановленной очередью на сохранение и глобальным помощником
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)


class LazyDelegateSavingPredefinedQueueFunction(
    LazySavingPredefinedQueueFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций с предустановленной очередью
    делегированным пусковику сохранением. В рамках функции производится
    наполнение очереди сохраняемых объектов, но сохранение будет производиться
    на уровне пусковика в открытой транзакции. Это требуется, когда функции
    должны выполняться цепочкой в рамках одной транзакции для достижения
    атомарности.
    """

    def do_save(self):
        """
        Выполнение сохранения объектов в зависимости от признака глобального
        сохранения создаваемых объектов функции.
        """
        self._do_save_objects_queue()

    def run(self):
        """
        Выполнение действий функции с дальнейшим сохранением объектов в базу
        при отсутствии ошибок и локальном сохранении
        """
        self._prepare()


class LazyDelegateSavingPredefinedQueueGlobalHelperFunction(
    GlobalHelperMixin,
    LazyDelegateSavingPredefinedQueueFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций с отложенным сохранением объектов,
    предустановленной очередью, делегированным сохранением объектов пускателю и
    глобальным помощником
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)


class LazySavingSettableQueueFunction(
    LazySavingFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций с отложенным сохранением объектов с
    устанавливаемой очередью объектов на сохранение
    """

    def __init__(
        self,
        *args,
        queue_to_save: Optional[Deque] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # очередь содержащая объекты на сохранение
        self._queue_to_save: Optional[Deque] = queue_to_save

    def set_queue_to_save(self, queue_to_save):
        """
        Установка очереди на сохранение
        """
        self._queue_to_save = queue_to_save


class LazySavingSettableQueueGlobalHelperFunction(
    GlobalHelperMixin,
    LazySavingSettableQueueFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций с отложенным сохранением объектов,
    устанавливаемой очередью и глобальным помощником
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)


class LazyDelegateSavingSettableQueueFunction(
    LazySavingSettableQueueFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций с устанавливаемой очередью
    делегированным пусковику сохранением. В рамках функции производится
    наполнение очереди сохраняемых объектов, но сохранение будет производиться
    на уровне пусковика в открытой транзакции. Это требуется, когда функции
    должны выполняться цепочкой в рамках одной транзакции для достижения
    атомарности. Используется в связке с LazySavingGeneralQueueRunner и его
    наследниками.
    """

    def do_save(self):
        """
        Сохранение делегировано на уровень пусковика
        """

    def run(self):
        """
        Выполнение действий функции с дальнейшим сохранением объектов
        """
        self._prepare()


class LazyDelegateSavingSettableQueueGlobalHelperFunction(
    GlobalHelperMixin,
    LazyDelegateSavingSettableQueueFunction,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания функций с отложенным сохранением объектов,
    устанавливаемой очередью, делегированным сохранением объектов пускателю и
    глобальным помощником. Используется в связке с LazySavingGeneralQueueRunner.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)



from django.apps import (
    AppConfig,
)

from function_tools.management.storages import (
    ImplementationStrategyStorage,
)


class FunctionToolsConfig(AppConfig):
    name = 'function_tools'
    label = 'function_tools'

    def ready(self):
        """
        На момент готовности приложения, необходимо добавить зарегистрированные стратегии в модель-перечисление
        ImplementationStrategy
        """
        from function_tools.models import (
            ImplementationStrategy,
        )
        strategy_storage = ImplementationStrategyStorage()

        for strategy in strategy_storage.implementation_strategy_map.values():
            if strategy.key:
                ImplementationStrategy.extend(
                    key=strategy.key.upper(),
                    title=strategy.title,
                )



from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import (
    Optional,
    Type,
)

from function_tools.runners import (
    BaseRunner,
)


class RunnerManager(metaclass=ABCMeta):
    """
    Абстрактный класс для создания менеджеров пусковиков запускаемых объектов.
    Служит для добавления функций-задач, пусковиков в очередь на исполнение.
    Является удобным механизмом для скрытия заполнения пусковика и его запуск.
    """
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        self._runner_class = self._prepare_runner_class()

        self._runner: Optional[BaseRunner] = None

    @property
    def result(self):
        return self._runner.result

    def _before_create_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера перед созданием ранера.
        """
    def _prepare_runner_class(self) -> Type[BaseRunner]:
        """
        Возвращает класс ранера.
        """
        return BaseRunner

    def _create_runner(self, *args, **kwargs):
        """
        Метод создания ранера.
        """
        if issubclass(self._runner_class, BaseRunner):
            self._runner = self._runner_class(*args, **kwargs)
        else:
            self._runner = BaseRunner(*args, **kwargs)

    def _after_create_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера после создания ранера.
        """

    def _before_prepare_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера перед подготовкой ранера.
        """

    @abstractmethod
    def _prepare_runner(self, *args, **kwargs):
        """
        Метод подготовки пусковика к запуску.

        Чаще всего используется для создания задач и заполнения ими очереди
        пусковика.
        """

    def _after_prepare_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера после подготовки ранера.
        """

    def _before_start_runner(self, *args, **kwargs):
        """
       Точка расширения поведения менеджера ранера перед запуском ранера.
       """

    def _start_runner(self, *args, **kwargs):
        """
        Точка расширения в месте запуска пусковика.
        """
        if self._runner:
            self._runner.run(*args, **kwargs)

    def _after_start_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера после запуска ранера.
        """

    def run(self, *args, **kwargs):
        self._before_create_runner(*args, **kwargs)
        self._create_runner(*args, **kwargs)
        self._after_create_runner(*args, **kwargs)

        self._before_prepare_runner(*args, **kwargs)
        self._prepare_runner(*args, **kwargs)
        self._after_prepare_runner(*args, **kwargs)

        self._before_start_runner(*args, **kwargs)
        self._start_runner(*args, **kwargs)
        self._after_start_runner(*args, **kwargs)


class LazySavingRunnerManager(RunnerManager):
    """
    Абстрактный класс для создания менеджеров пусковиков c отложенным
    сохранением. В качестве пусковика должен использоваться потомок
    LazySavingRunner.
    """

    def _before_do_save(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера перед запуском сохранения объектов.
        """

    def _do_save(self, *args, **kwargs):
        """
        Запуск сохранения пусковика
        """
        self._runner.do_save(*args, **kwargs)

    def _after_do_save(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера после запуска сохранения объектов.
        """

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

        self._before_do_save(*args, **kwargs)
        self._do_save(*args, **kwargs)
        self._after_do_save(*args, **kwargs)



class BaseValidator:
    """
    Базовый класс для создания валидатора
    """

    def __init__(self, *args, **kwargs):
        super().__init__()

    def validate(self, runnable):
        """
        Точка входа работы валидатора.

        Результат валидации должен быть записан в runnable.result
        """



from typing import (
    List,
)


class BaseError:
    """
    Базовый класс для создания возникающих ошибок для дальнейшей обработки и
    вывода
    """

    def __init__(
        self,
        message: str = '',
    ):
        self._message = message

    def as_list(self) -> List[str]:
        return [self._message]

    def as_str(self) -> str:
        return self._message



SEARCHING_KEY_SIZE_MORE_THAN_DEFAULT_SEARCHING_KEY_ERROR = (
    'Ключ для поиска не может быть больше, чем ключ, на основе которого была '
    'построена хеш-таблица!'
)

SEARCHING_KEY_SIZE_LESS_THAN_DEFAULT_SEARCHING_KEY_ERROR = (
    'Ключ для поиска не может быть меньше, чем ключ, на основе которого была '
    'построена хеш-таблица в строгом режиме поиска!'
)

DATE_FROM_MORE_OR_EQUAL_DATE_TO_ERROR = (
    'Начальная дата должна быть строго меньше конечной даты'
)




from django.contrib.postgres.fields import (
    ArrayField,
)
from django.db.models import (
    CharField,
    DateTimeField,
    Model,
)

from m3_db_utils.models import (
    TitledModelEnum,
)


class ImplementationStrategy(TitledModelEnum):
    """
    Перечисление стратегий реализации функций
    """

    class Meta:
        db_table = 'function_tools_implementation_strategy'
        verbose_name = 'Стратегия создания функции'
        extensible = True


class RegisteredFunction(Model):
    """
    Модель зарегистрированных функций системы
    """

    function_path = CharField(
        verbose_name='Путь до функции',
        primary_key=True,
        db_index=True,
        unique=True,
        max_length=1024,
    )
    verbose_name = CharField(
        verbose_name='Название функции',
        max_length=512,
        default='Имя функции не определено',
    )

    tags = ArrayField(base_field=CharField(max_length=128, blank=True), default=list)

    created_at = DateTimeField(
        verbose_name='Дата и время создания',
        auto_now_add=True,
    )
    updated_at = DateTimeField(
        verbose_name='Дата и время обновления',
        auto_now=True,
    )

    class Meta:
        db_table = 'function_tools_registered_function'
        verbose_name = 'Зарегистрированная функция'



from abc import (
    ABCMeta,
)
from collections import (
    deque,
)
from typing import (
    Deque,
    Optional,
    Tuple,
    Union,
)

from function_tools.errors import (
    BaseError,
)
from function_tools.functions import (
    LazyDelegateSavingSettableQueueFunction,
)
from function_tools.general import (
    LazySavingActionModelRunnableObject,
    LazySavingRunnableObject,
    RunnableObject,
)
from function_tools.helpers import (
    BaseHelper,
    BaseRunnerHelper,
)
from function_tools.mixins import (
    GlobalHelperMixin,
    HelperMixin,
)


class BaseRunner(
    HelperMixin,
    RunnableObject,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания пусковиков выполнения запускаемых объектов
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._helper: Union[BaseHelper, BaseRunnerHelper] = self._prepare_helper(*args, **kwargs)
        self._queue: Deque[RunnableObject] = deque()

    def _prepare_runnable_before_enqueue(
        self,
        runnable: RunnableObject,
    ):
        """
        Подготовка запускаемого объекта к работе.

        В данной точке расширения можно пропатчить объект через публичные методы
        """
        if isinstance(runnable, GlobalHelperMixin):
            runnable.set_global_helper(
                global_helper=self._helper,
            )

    def enqueue(
        self,
        runnable: Union[RunnableObject, Tuple[RunnableObject, ...]],
        *args,
        **kwargs,
    ):
        """
        Добавление задачи на выполнение функции в очередь
        """
        if not isinstance(runnable, tuple):
            runnable = (runnable, )

        for r in runnable:
            self._prepare_runnable_before_enqueue(
                runnable=r,
            )

            self._queue.append(r)

    def _update_queue(self):
        """Метод позволяет обновлять очередь ранеров на ходу"""
        pass

    def run(self, *args, **kwargs):
        """
        Выполнение всех задач стоящих в очереди
        """
        self.before_validate()
        self.validate()
        self.after_validate()

        if self.result.has_not_errors:
            # атрибут отвечающий за передачу результата следующему в цепочке ранеру
            prev_runnable_result = None

            self._update_queue()

            while self._queue:
                runnable: RunnableObject = (
                    self._queue.popleft()
                )

                runnable.adopt_prev_runnable_result(prev_runnable_result=prev_runnable_result)

                runnable.before_validate()
                runnable.validate()
                runnable.after_validate()

                runnable.run()

                prev_runnable_result = runnable.result

                self.result.append_entity(runnable.result)


class LazySavingRunner(
    BaseRunner,
    LazySavingRunnableObject,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания классов пусковиков с отложенным сохранением
    объектов из очередей на сохранение запускаемых объектов.

    Сохранение производится, когда все запускаемые объекты очереди отработают.
    """

    def _do_save_objects_queue(self):
        """
        Запуск сохранения у выполняемых объектов
        """
        while self._queue_to_save:
            runnable: LazySavingRunnableObject = self._queue_to_save.popleft()
            runnable.do_save()

    def run(self, *args, **kwargs):
        """
        Выполнение всех задач стоящих в очереди
        """
        self.before_validate()
        self.validate()
        self.after_validate()

        if self.result.has_not_errors:
            # атрибут отвечающий за передачу результата следующему в цепочке ранеру
            prev_runnable_result = None

            self._update_queue()

            while self._queue:
                runnable: RunnableObject = self._queue.popleft()

                runnable.adopt_prev_runnable_result(prev_runnable_result=prev_runnable_result)

                runnable.before_validate()
                runnable.validate()
                runnable.after_validate()

                runnable.run()

                prev_runnable_result = runnable.result

                self.result.append_entity(runnable.result)

                if runnable.result.has_not_errors:
                    self._queue_to_save.append(runnable)


class LazyStrictSavingRunner(
    LazySavingRunner,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания классов пусковиков с отложенным сохранением
    объектов в строгом режиме.

    Если не все выполняемые объекты отработали корректно, то ни один не
    сохраняется.
    """

    def _get_strict_saving_error(self) -> BaseError:
        """
        Ошибка, которая должна быть возвращена при несоблюдении условий
        строгого режима
        """
        return BaseError()

    def run(self, *args, **kwargs):
        """
        Выполнение всех задач стоящих в очереди
        """
        self.before_validate()
        self.validate()
        self.after_validate()

        if self.result.has_not_errors:
            prev_runnable_result = None
            queue_length = len(self._queue)

            self._update_queue()

            while self._queue:
                runnable: RunnableObject = (
                    self._queue.popleft()
                )

                runnable.adopt_prev_runnable_result(prev_runnable_result=prev_runnable_result)

                runnable.before_validate()
                runnable.validate()
                runnable.after_validate()

                runnable.run()

                prev_runnable_result = runnable.result

                self.result.append_entity(runnable.result)

                if runnable.result.has_not_errors:
                    self._queue_to_save.append(runnable)

            if queue_length != len(self._queue_to_save):
                self.result.append_entity(
                    self._get_strict_saving_error()
                )

                self._queue_to_save.clear()


class GlobalHelperRunner(
    GlobalHelperMixin,
    BaseRunner,
    metaclass=ABCMeta,
):
    """
    Базовый класс для создания пусковиков выполнения запускаемых объектов с
    глобальным помощником
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)


class LazySavingGlobalHelperRunner(
    GlobalHelperMixin,
    LazySavingRunner,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания классов пусковиков с отложенным сохранением
    объектов с глобальным помощником
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)


class LazyStrictSavingGlobalHelperRunner(
    GlobalHelperMixin,
    LazyStrictSavingRunner,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания классов пусковиков с отложенным сохранением объектов в строгом режиме c
    глобальным помощником
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._global_helper: BaseRunnerHelper = self._prepare_global_helper(*args, **kwargs)


class LazySavingSettableQueueRunner(
    BaseRunner,
    LazySavingRunnableObject,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания пусковиков с устанавливаемой очередью на
    сохранение.
    """

    def __init__(
        self,
        *args,
        queue_to_save: Optional[Deque] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # очередь содержащая объекты на сохранение
        self._queue_to_save: Optional[Deque] = queue_to_save

    def set_queue(self, queue_to_save):
        """
        Установка очереди на сохранение
        """
        self._queue_to_save = queue_to_save

    def _prepare_runnable_before_enqueue(
        self,
        runnable: Union['LazySavingSettableQueueRunner', LazyDelegateSavingSettableQueueFunction],  # noqa
    ):
        runnable.set_queue_to_save(
            queue_to_save=self._queue_to_save,
        )


class LazyDelegateSavingSettableQueueRunner(
    LazySavingSettableQueueRunner,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания пусковиков с устанавливаемой очередью на
    сохранение. Используется в связке с LazySavingGeneralQueueRunner в качестве
    пусковкика. В качестве запускаемых объектов используются
    LazyDelegateSavingSettableQueueFunction и его потомки, и объекты самого
    класса и его потомков.
    """

    def do_save(self, *args, **kwargs):
        """
        Сохранение делегировано пусковику
        """


class LazySavingGeneralQueueRunner(
    BaseRunner,
    LazySavingActionModelRunnableObject,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания пусковиков с единой очередью сохранения для
    всех исполняемых объектов. Используется в паре
    LazyDelegateSavingSettableQueueFunction,
    LazyDelegateSavingSettableQueueRunner и их наследниками.
    """

    def _prepare_runnable_before_enqueue(
        self,
        runnable: Union['LazyDelegateSavingSettableQueueRunner', LazyDelegateSavingSettableQueueFunction],  # noqa
    ):
        runnable.set_queue_to_save(
            queue_to_save=self._queue_to_save,
        )



# Наименование модуля с функцией
FUNCTIONS_MODULE = 'functions'

# Путь модуля с базовыми функциями
FUNCTION_TOOLS_FUNCTIONS_MODULE_PATH = f'function_tools.{FUNCTIONS_MODULE}'

# Пустое значение полного наименования функции
EMPTY_VERBOSE_NAME = 'Empty verbose name'



from abc import (
    ABCMeta,
    abstractmethod,
)

from function_tools.results import (
    BaseRunnableResult,
)


class ResultPresenter(metaclass=ABCMeta):
    """
    Презентер результата выполнения запускаемого объекта.
    """
    def __init__(self, runnable_result: BaseRunnableResult):
        self._runnable_result = runnable_result

        self._presentable_result = None

    @abstractmethod
    def represent(self):
        """
        Осуществляется преобразование результата выполнения запускаемого
        объекта в нужный вид
        """






import inspect
import sys
from importlib import (
    import_module,
)
from importlib._bootstrap import (
    module_from_spec,
)
from importlib._bootstrap_external import (
    spec_from_file_location,
)
from pathlib import (
    Path,
)
from typing import (
    Dict,
    List,
    Tuple,
    Type,
)

from django.conf import (
    settings,
)

from function_tools.management.strategies import (
    FunctionImplementationStrategy,
)


class ImplementationStrategyStorage:
    """
    Хранилище реализованных стратегий создания функций
    """

    def __init__(self):
        self._implementation_strategy_map = self._prepare_implementation_strategy_map()

    @property
    def implementation_strategy_map(self) -> Dict[str, Type[FunctionImplementationStrategy]]:
        """
        Возвращает карту соответствия зарегистрированных стратегий реализации функций
        """
        return self._implementation_strategy_map

    def get_strategy_implementation(
        self,
        strategy_key: str,
    ):
        """
        Возвращает стратегию реализации функции по значению перечисления
        """
        return self._implementation_strategy_map[strategy_key]

    def _find_strategies_modules(self) -> List[Tuple[str, str]]:
        """
        Поиск модулей содержащих стратегии создания функций
        """
        strategies_modules = []
        checked_packed_paths = []

        for app_name in settings.INSTALLED_APPS:
            # TODO EDUSCHL-17934 В app_name могут прилетать AppConfig. Необходимо произвести доработку и обработать
            #  такие случаи.
            try:
                app_module = import_module(app_name)
            except ModuleNotFoundError as e:
                sys.stdout.write(str(e))
                continue

            app_path = app_module.__path__

            # Если поиск уже осуществлялся по родительской директории,
            # то проверку нужно пропустить
            is_already_checked = False
            for check_packed_path in checked_packed_paths:
                if app_path in check_packed_path:
                    is_already_checked = True

                    break

            if is_already_checked:
                continue

            application_path = Path(app_path[0])
            strategies_file_patterns = [
                '**/**/strategies.py',
            ]

            for strategies_file_pattern in strategies_file_patterns:
                strategies_modules_paths = application_path.glob(
                    strategies_file_pattern
                )
                for strategies_module_path in strategies_modules_paths:
                    strategies_module_path = str(strategies_module_path)

                    module_name = (
                        str(strategies_module_path).split('/')[-1].split('.')[0]
                    )

                    spec = spec_from_file_location(
                        name=module_name,
                        location=strategies_module_path,
                    )
                    strategies_module = module_from_spec(
                        spec=spec,
                    )
                    spec.loader.exec_module(strategies_module)
                    strategies_modules.append(
                        (
                            strategies_module,
                            strategies_module_path,
                        )
                    )

        return strategies_modules

    def _prepare_implementation_strategy_map(self) -> Dict[str, FunctionImplementationStrategy]:
        """
        Поиск фабрик во всех подключенных приложениях
        """
        strategy_map = {}
        strategies_modules = self._find_strategies_modules()

        for strategies_module, module_path in strategies_modules:
            strategy_classes_names = list(filter(
                lambda name: (
                    inspect.isclass(getattr(strategies_module, name))
                    # TODO Следуем соглашению, что имя класса стратегии реализации функции заканчивается
                    #  ImplementationStrategy. issubclass не удается использовать из за возникновения ошибок после
                    #  импорта базовой стратегии
                    and getattr(strategies_module, name).__name__.endswith('ImplementationStrategy')
                    and not inspect.isabstract(getattr(strategies_module, name))
                ),
                dir(strategies_module)
            ))

            if strategy_classes_names:
                for strategy_class_name in strategy_classes_names:
                    strategy_class: Type[FunctionImplementationStrategy] = getattr(
                        strategies_module,
                        strategy_class_name
                    )

                    strategy = strategy_class()

                    if strategy.key in strategy_map:
                        if strategy.__class__.__name__ != strategy_map[strategy.key].__class__.__name__:
                            sys.stdout.write(
                                f'Please, check {strategy.__class__.__name__} and '
                                f'{strategy_map[strategy.key].__class__.__name__} implementation. Found duplicates '
                                f'unique strategies implementation key {strategy.key}\n'
                            )
                    else:
                        strategy_map[strategy.key] = strategy

        return strategy_map



# Generated by Django 3.2.6 on 2022-02-16 13:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('function_tools', '0003_implementationstrategy'),
    ]

    operations = [
        migrations.AddField(
            model_name='implementationstrategy',
            name='order_number',
            field=models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер'),
        ),
    ]



from codecs import (
    open,
)
from os import (
    path,
)

from setuptools import (
    find_packages,
    setup,
)


__version__ = '0.2.3'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'CHANGES.md'), encoding='utf-8') as f:
    long_description += f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [
    x.strip()
    for x in all_reqs if
    'git+' not in x
]
dependency_links = [
    x.strip().replace('git+', '')
    for x in all_reqs if
    x.startswith('git+')
]

excluded_packages = (
    'docs',
)

setup(
    name='function-tools',
    version=__version__,
    description='Tools for creating functions',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='',
    packages=find_packages(exclude=excluded_packages),
    include_package_data=True,
    author='Alexander Danilenko',
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email='a.danilenko@bars.group',
    url='https://github.com/sandanilenko/function-tools',
)



from collections import (
    Iterable,
    OrderedDict,
    Sequence,
)
from copy import (
    deepcopy,
)
from datetime import (
    date,
)
from operator import (
    attrgetter,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from django.db.models import (
    Model,
)

from function_tools.enums import (
    TransferPeriodEnum,
)
from function_tools.strings import (
    DATE_FROM_MORE_OR_EQUAL_DATE_TO_ERROR,
    SEARCHING_KEY_SIZE_LESS_THAN_DEFAULT_SEARCHING_KEY_ERROR,
    SEARCHING_KEY_SIZE_MORE_THAN_DEFAULT_SEARCHING_KEY_ERROR,
)
from function_tools.types import (
    QueryType,
)
from function_tools.utils import (
    date2str,
    deep_getattr,
)


class BaseCache:
    """
    Кеш-заглушка
    """

    def __init__(self, *args, **kwargs):
        super().__init__()


class EntityCache(BaseCache):
    """
    Базовый класс кеша объектов сущности
    """

    def __init__(
        self,
        model: Type[Model],
        *args,
        select_related_fields: Optional[Tuple[str, ...]] = None,
        only_fields: Optional[Tuple[str, ...]] = None,
        additional_filter_params: Optional[Dict[str, Any]] = None,
        searching_key: Union[str, Tuple[str, ...]] = ('pk', ),
        **kwargs,

    ):
        super().__init__(*args, **kwargs)

        self._model = model
        self._select_related_fields = select_related_fields
        self._only_fields = only_fields
        self._actual_entities_queryset = self._prepare_actual_entities_queryset()  # noqa

        self._searching_key = (
            searching_key if (
                isinstance(searching_key, Iterable) and
                not isinstance(searching_key, str)
            ) else
            (searching_key, )
        )

        self._additional_filter_params = additional_filter_params or {}

        self._entities = None
        self._entities_hash_table = None

        self._before_prepare()
        self._prepare()
        self._after_prepare()

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} @model="{self._model.__name__}" '
            f'@searching_key="{self._searching_key}">'
        )

    def __str__(self):
        return self.__repr__()

    @property
    def entities(self) -> Union[QueryType[Model], List[Model]]:
        """
        Возвращает список объектов либо QuerySet в зависимости от подхода к
        получению объектов
        """
        return self._entities

    def _before_prepare(self):
        """
        Точка расширения перед подготовкой кеша
        """
        pass

    def _after_prepare(self):
        """
        Точка расширения после подготовки кеша
        """
        pass

    def _prepare(self):
        """
        Метод подготовки кеша
        """
        self._prepare_entities()
        self._prepare_entities_hash_table()

    def _prepare_entities(self):
        """
        Получение выборки объектов модели по указанными параметрам
        """
        self._entities = self._actual_entities_queryset.filter(
            **self._additional_filter_params,
        )

        if self._only_fields:
            self._entities = self._entities.only(*self._only_fields)

        self._entities = self._entities.distinct()

    def _prepare_entities_hash_table(self):
        """
        Отвечает за построение хеш таблицы для дальнейшего поиска. В качестве
        ключа можно задавать строку - наименование поля или кортеж наименований
        полей.

        Если требуется доступ через внешний ключ, то необходимо использовать
        точку в качестве разделителей. Например,
        searching_key = tuple('account_id', 'supplier.code')
        """
        hash_table = {}

        key_items_count = len(self._searching_key)
        for entity in self._entities:
            temp_hash_item = hash_table

            for index, key_item in enumerate(self._searching_key, start=1):
                key_item_value = deep_getattr(entity, key_item)

                if key_item_value is not None:
                    if index == key_items_count:
                        if key_item_value not in temp_hash_item:
                            temp_hash_item[key_item_value] = entity
                        else:
                            if isinstance(temp_hash_item[key_item_value], set):
                                temp_hash_item[key_item_value].add(
                                    entity
                                )
                            else:
                                temp_hash_item[key_item_value] = {
                                    temp_hash_item[key_item_value], entity
                                }
                    else:
                        if key_item_value not in temp_hash_item:
                            temp_hash_item[key_item_value] = {}

                        temp_hash_item = temp_hash_item[key_item_value]
                else:
                    break

        self._entities_hash_table = hash_table

    def _prepare_actual_entities_queryset(self):
        """
        Подготовка менеджера с указанием идентификатора учреждения и состояния,
        если такие имеются у модели
        """
        actual_entities_queryset = self._model._base_manager.all()

        if self._select_related_fields:
            actual_entities_queryset = actual_entities_queryset.select_related(
                *self._select_related_fields,
            )

        return actual_entities_queryset

    def filter(
        self,
        only_first: bool = False,
        **kwargs,
    ):
        """
        Метод фильтрации объектов кеша по заданным параметрам.

        Пример использования:

        some_objects_list = cache.filter(code='12345', only_first=True)

        Можно получать первое попавшееся значение с указанием only_first=True
        """
        filter_ = OrderedDict()
        for field_name, field_value in kwargs.items():
            prepared_field_name = field_name.split('__in')[0]
            prepared_field_value = (
                set(field_value) if
                self._check_is_iterable(field_value) else
                {field_value}
            )
            filter_[prepared_field_name] = prepared_field_value

        result = list(
            filter(
                lambda entity: all(
                    [
                        getattr(entity, field_name) in filter_[field_name]
                        for field_name in filter_.keys()
                    ]
                ),
                self._entities
            )
        )

        if only_first:
            if result:
                result = result[0]
            else:
                result = None

        return result

    def flat_values_list(
        self,
        field_name: str,
    ):
        """
        Получение плоского списка значений объектов указанного свойства без
        пустых значений.

        : param field_name: наименование поля
        """
        return list(
            filter(
                None,
                [
                    getattr(entity, field_name, None)
                    for entity in self._entities
                ]
            )
        )

    def _check_is_iterable(self, object_):
        """
        Проверяет, является ли передаваемый объект итерабельным
        """
        return (
            isinstance(object_, (Iterable, Sequence)) and
            not isinstance(object_, str)
        )

    def get_by_key(
        self,
        key: Union[Any, Tuple[Any]],
        strict_mode=True,
    ):
        """
        Метод получения значения из кеша по ключу поиска.

        В общем случае передаваемый ключ должен совпадать с _searching_key.
        Если отключить строгий режим поиска - strict_mode=False, то можно
        получить промежуточный результат по части ключа следующего с начала.
        """
        key = (
            key if
            self._check_is_iterable(key) else
            (key, )
        )
        key_items_count = len(key)
        searching_key_items_count = len(self._searching_key)

        if key_items_count > searching_key_items_count:
            raise ValueError(
                SEARCHING_KEY_SIZE_MORE_THAN_DEFAULT_SEARCHING_KEY_ERROR
            )
        elif key_items_count < searching_key_items_count and strict_mode:
            raise ValueError(
                SEARCHING_KEY_SIZE_LESS_THAN_DEFAULT_SEARCHING_KEY_ERROR
            )

        result = None
        temp_hash_item = self._entities_hash_table
        for index, key_item in enumerate(key, start=1):
            if key_item not in temp_hash_item:
                if not strict_mode and index != 1:
                    result = temp_hash_item

                break

            temp_hash_item = temp_hash_item[key_item]

            if index == key_items_count:
                result = temp_hash_item

        return result

    def values_list(
        self,
        fields: Tuple[str, ...],
    ) -> Optional[List[Tuple]]:
        """
        Получение списка кортежей состоящих из значений полей объектов согласно
        заданным параметрам

        :param fields: кортеж наименований полей
        """
        fields_getter = attrgetter(*fields)

        return [
            fields_getter(entity)
            for entity in self._entities
        ]

    def first(self):
        """
        Получение первого элемента из кеша
        """
        result = None

        if self._entities:
            result = self._entities[0]

        return result


class ActualEntityCache(EntityCache):
    """
    Базовый класс кеша актуальных записей сущности
    """

    def __init__(
        self,
        actual_date,
        *args,
        **kwargs,
    ):
        self._actual_date = actual_date

        super().__init__(*args, **kwargs)

    def _prepare_actual_entities_queryset(self) -> Dict[str, date]:
        """
        Метод получения фильтра актуализации по дате.
        """
        actual_entities_queryset = super()._prepare_actual_entities_queryset()

        actual_entities_queryset = actual_entities_queryset.filter(
            begin__lte=self._actual_date,
            end__gte=self._actual_date,
        )

        return actual_entities_queryset


class PeriodicalEntityCache(BaseCache):
    """
    Базовый класс периодического кеша.

    Кеш создается для определенной модели с указанием двух дат, на которые
    должны быть собраны кеши актуальных объектов модели.

    Для примера, может использоваться при переносах остатков на очередной год
    с 31 декабря на 1 января нового года.
    """

    entity_cache_class = EntityCache

    def __init__(
        self,
        date_from: date,
        date_to: date,
        model: Type[Model],
        *args,
        select_related_fields: Optional[Tuple[str, ...]] = None,
        only_fields: Optional[Tuple[str, ...]] = None,
        additional_filter_params: Optional[Dict[str, Any]] = None,
        searching_key: Union[str, Tuple[str, ...]] = ('pk', ),
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if date_from >= date_to:
            raise ValueError(
                DATE_FROM_MORE_OR_EQUAL_DATE_TO_ERROR
            )

        self._model = model
        self._select_related_fields = select_related_fields
        self._only_fields = only_fields

        self._additional_filter_params = additional_filter_params or {}
        self._searching_key = searching_key

        self._date_from = date_from
        self._date_to = date_to

        self._old_entities_cache: Optional[EntityCache] = None
        self._new_entities_cache: Optional[EntityCache] = None

        self._before_prepare()
        self._prepare()
        self._after_prepare()

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} @model="{self._model.__name__}" '
            f'@date_from="{date2str(self._date_from)}" '
            f'@date_to="{date2str(self._date_to)}" '
            f'@searching_key="{self._searching_key}">'
        )

    def __str__(self):
        return self.__repr__()

    @property
    def old(self):
        """
        Кеш объектов модели актуальных на начальную дату
        """
        return self._old_entities_cache

    @property
    def new(self):
        """
        Кеш объектов модели актуальный на конечную дату
        """
        return self._new_entities_cache

    def _get_actuality_filter(
        self,
        period_type: str,
    ) -> Dict[str, date]:
        """
        Метод получения фильтра актуализации по дате.

        При получении счетов или аналитик при переносе остатков необходимо
        учитывать период действия следуя следующей логике:
        -- старые - begin < date_from &&  end >= date_from
        -- новые - begin <= date_to && end > date_to

        :param dict period_type: словарь с параметрами для актуализации по дате
        :return:
        """
        if period_type == TransferPeriodEnum.OLD:
            actuality_filter = {
                'begin__lt': self._date_from,
                'end__gte': self._date_from,
            }
        else:
            actuality_filter = {
                'begin__lte': self._date_to,
                'end__gt': self._date_to,
            }

        return actuality_filter

    def _prepare_entities_cache(
        self,
        additional_filter_params: Optional[Dict[str, Any]]
    ):
        """
        Создание кеша объектов модели на указанную дату по указанным параметром
        с ключом поиска для построения хеш-таблицы
        """
        entities_cache = self.entity_cache_class(
            model=self._model,
            select_related_fields=self._select_related_fields,
            only_fields=self._only_fields,
            additional_filter_params=additional_filter_params,
            searching_key=self._searching_key,
        )

        return entities_cache

    def _prepare_additional_filter_params(
        self,
        period_type: str,
    ):
        """
        Подготовка словаря с дополнительными параметрами для дальнейшей
        фильтрации объектов при формировании кеша
        """
        additional_filter_params = deepcopy(self._additional_filter_params)
        additional_filter_params.update(
            **self._get_actuality_filter(
                period_type=period_type,
            )
        )

        return additional_filter_params

    def _prepare_old_additional_filter_params(self):
        """
        Подготовка дополнительных параметров фильтрации на начальную дату
        """
        return self._prepare_additional_filter_params(
            period_type=TransferPeriodEnum.OLD,
        )

    def _prepare_new_additional_filter_params(self):
        """
        Подготовка дополнительных параметров фильтрации на конечную дату
        """
        return self._prepare_additional_filter_params(
            period_type=TransferPeriodEnum.NEW,
        )

    def _prepare_old_entities_cache(self):
        """
        Формирование кеша объектов модели на начальную дату
        """
        additional_filter_params = self._prepare_old_additional_filter_params()
        self._old_entities_cache = self._prepare_entities_cache(
            additional_filter_params=additional_filter_params,
        )

    def _prepare_new_entities_cache(self):
        """
        Формирование кеша объектов модели на конечную дату
        """
        additional_filter_params = self._prepare_new_additional_filter_params()
        self._new_entities_cache = self._prepare_entities_cache(
            additional_filter_params=additional_filter_params,
        )

    def _before_prepare(self):
        """
        Точка расширения перед формированием кеша
        """

    def _prepare(self):
        """
        Формирование кешей на начальную и конечную даты
        """
        self._prepare_old_entities_cache()
        self._prepare_new_entities_cache()

    def _after_prepare(self):
        """
        Точка расширения после формирования кеша
        """


class CacheStorage(BaseCache):
    """
    Хранилище кешей.

    Для выполнения функций, в большинстве случаев, необходимы кеши для
    множества сущностей созданные по особым правилам, но подчиняющиеся общим.
    Для их объединения и применения в функции создаются хранилища, содержащие кеши в виде публичных свойств, с
    которыми в дальнейшем удобно работать
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import sys
from importlib import (
    import_module,
)
from inspect import (
    isclass,
)
from typing import (
    Dict,
    List,
    Tuple,
)

from django.apps import (
    apps,
)
from django.core.management import (
    color_style,
)
from django.db import (
    ProgrammingError,
)
from django.db.models.signals import (
    post_migrate,
)

from function_tools.consts import (
    FUNCTION_TOOLS_FUNCTIONS_MODULE_PATH,
    FUNCTIONS_MODULE,
)
from function_tools.functions import (
    BaseFunction,
)
from function_tools.models import (
    RegisteredFunction,
)


class FunctionRegistrar:
    """
    Регистратор функций системы

    Отвечает за поиск всех реализованных функций системы и их регистрацию в модели RegisteredFunction. Путь до
    приложения функции является уникальным, т.к. каждая функция реализуется в отдельном приложении и это приложение
    регистрируется в INSTALLED_APPS. Также указывается название класса функции и ее наименование. Имя функции берется
    из конфига Django-приложения из свойства verbose_name
    """

    def __init__(self):
        self._functions: Dict[str, Tuple[str, List[str]]] = {}

        self._registered_functions = RegisteredFunction.objects.all()
        self._registered_functions_map = {
            registered_function.function_path: registered_function
            for registered_function in self._registered_functions
        }

        self._style = color_style()

    def _find_functions(self):
        """
        Поиск реализованных функций системы
        """
        functions_module = import_module(FUNCTION_TOOLS_FUNCTIONS_MODULE_PATH)
        excluded_base_functions = [
            name
            for name, class_ in functions_module.__dict__.items()
            if isclass(class_) and issubclass(class_, BaseFunction)
        ]

        app_configs = apps.get_app_configs()
        function_app_configs = list(filter(lambda config: getattr(config.module, FUNCTIONS_MODULE, None), app_configs))

        for app_config in function_app_configs:
            app_functions = [
                (name, class_)
                for name, class_ in app_config.module.functions.__dict__.items()
                if isclass(class_) and issubclass(class_, BaseFunction) and name not in excluded_base_functions
            ]

            for app_function in app_functions:
                self._functions[f'{app_config.module.__name__}.functions.{app_function[0]}'] = (
                    app_function[1].verbose_name, app_function[1].tags
                )

    def _process_creating_functions(self):
        """
        Обработка создаваемых функций системы
        """
        for_creating = []

        creating_function_paths = filter(lambda f_p: f_p not in self._registered_functions_map, self._functions.keys())

        for creating_function_path in creating_function_paths:
            for_creating.append(
                RegisteredFunction(
                    function_path=creating_function_path,
                    verbose_name=self._functions[creating_function_path][0],
                    tags=self._functions[creating_function_path][1],
                )
            )

        RegisteredFunction.objects.bulk_create(objs=for_creating)

    def _process_updating_functions(self):
        """
        Обработка обновляемых функций системы
        """
        updating_function_paths = filter(lambda f_p: f_p in self._registered_functions_map, self._functions.keys())

        for updating_function_path in updating_function_paths:
            if (
                self._registered_functions_map[updating_function_path].verbose_name != self._functions[updating_function_path][0]  # noqa
                or self._registered_functions_map[updating_function_path].tags != self._functions[updating_function_path][1]  # noqa
            ):
                self._registered_functions_map[updating_function_path].verbose_name = self._functions[updating_function_path][0]  # noqa
                self._registered_functions_map[updating_function_path].tags = self._functions[updating_function_path][1]  # noqa

                # Иду на преступный шаг с генерацией некоторого количества запросов в БД, все только для того, чтобы
                # изменить дату и время обновления записи
                self._registered_functions_map[updating_function_path].save()

    def _process_deleted_functions(self):
        """
        Обработка удаленных, перенесенных или находящихся в отключенных плагинах функций системы
        """
        deleted_functions = set(self._registered_functions_map.keys()).difference(self._functions.keys())
        if deleted_functions:
            for deleted_function in deleted_functions:
                sys.stdout.write(
                    self._style.NOTICE(
                        f'Found registered function in the database, but not found implementation. Function with path "'
                        f'{deleted_function}". Please, check it!\n'
                    )
                )

    def _register(self):
        """
        Регистрация реализованных функций системы
        """
        self._process_creating_functions()
        self._process_updating_functions()
        self._process_deleted_functions()

    def run(self):
        self._find_functions()
        self._register()


def post_migrate_receiver(sender, app_config, **kwargs):
    """
    Выполняет регистрацию реализованных в проекте функций
    """
    if app_config.name == 'function_tools':
        try:
            registrar = FunctionRegistrar()
            registrar.run()
        except ProgrammingError as e:
            # TODO EDUSCHL-17935 При первом прогоне миграций может возникать ошибка, т.к. транзакция еще не
            #  зафиксирована. Необходимо найти более оптимальное решение.
            sys.stdout.write(str(e))


post_migrate.connect(post_migrate_receiver)



from abc import (
    ABCMeta,
    abstractmethod,
)
from collections import (
    deque,
)
from typing import (
    Deque,
    Type,
)

from django.db import (
    transaction,
)
from django.db.models import (
    Model,
)

from function_tools.mixins import (
    ValidatorMixin,
)
from function_tools.results import (
    BaseRunnableResult,
)
from function_tools.utils import (
    rebind_model_rel_id,
)
from function_tools.validators import (
    BaseValidator,
)


class RunnableObject(
    ValidatorMixin,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания классов выполняемых объектов
    """

    def __init__(self, *args, **kwargs):
        self._validator: BaseValidator = self._prepare_validator(*args, **kwargs)
        self._result: BaseRunnableResult = self._prepare_result(*args, **kwargs)

    def _prepare_result_class(self) -> Type[BaseRunnableResult]:
        """
        Возвращает класс результата работы ранера
        """
        return BaseRunnableResult

    @property
    def result(self) -> BaseRunnableResult:
        return self._result

    def _prepare_result(self, *args, **kwargs):
        """
        Метод подготовки результата
        """
        result_class = self._prepare_result_class()

        if issubclass(result_class, BaseRunnableResult):
            result = result_class(*args, **kwargs)
        else:
            result = BaseRunnableResult(*args, **kwargs)

        return result

    def adopt_prev_runnable_result(self, prev_runnable_result):
        """Метод перенимания результата предыдущего ранера."""
        pass

    @abstractmethod
    def run(self, *args, **kwargs):
        """
        Метод запуска выполняемого объекта
        """


class LazySavingRunnableObject(
    RunnableObject,
    metaclass=ABCMeta,
):
    """
    Абстрактный класс для создания классов с отложенным сохранением

    Функции или другие исполняемые объекты должны иметь возможность быть
    атомарными. Для достижения этой цели необходимо все
    действия по сохранению выполнять в транзакции. Если сохранение
    объектов моделей производится непосредственно в коде, то возникает
    ситуация, когда транзакции становятся длинными. По логике вещей, в
    транзанзакцию должны быть обернуты действия на сохранение объектов. Это
    достигается с помощью предоставляемого функционала. В процессе работы
    функции не производится сохранение объектов модели, а они складываются в
    очередь на сохранение. Создаваемые объекты используются в качестве внешних
    объектов, как и ранее. После выполнения всех действий вызывается метод
    do_save, который открывает транзакцию и выполняет все запланированные на
    выполнение действия в порядке добавления в очередь.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._queue_to_save: Deque = deque()

    def do_on_save(self, object_):
        """
        Метод добавления объекта в очередь на сохранение
        """
        self._queue_to_save.append(object_)

    @abstractmethod
    def _do_save_objects_queue(self):
        """
        Выполенение сохранения объектов из очереди
        """

    def do_save(self):
        """
        Выполнение действий сохранения объектов из очереди в транзакции
        """
        with transaction.atomic(savepoint=False):
            self._do_save_objects_queue()


class LazySavingActionModelRunnableObject(
    LazySavingRunnableObject,
    metaclass=ABCMeta,
):
    """
    Абстрактный для создания классов исполняемых объектов с отложенным сохранением, использующих в качестве объектов
    объекты моделей и функции как действия. Обычно функции - это замыкания.
    """
    def _do_save_objects_queue(self):
        """
        Выполенение сохранения объектов из очереди
        """
        while self._queue_to_save:
            x = self._queue_to_save.popleft()

            if callable(x):
                x()
            else:
                if isinstance(x, Model):
                    rebind_model_rel_id(x)
                    x.save()



import operator

from django.conf import (
    settings,
)
from django.core.exceptions import (
    ObjectDoesNotExist,
)
from django.db.models import (
    Model,
)
from django.utils import (
    datetime_safe,
)


def deep_getattr(
    obj,
    attr,
    default=None,
):
    """
    Получить значение атрибута с любого уровня цепочки вложенных объектов.

    :param object obj: объект, у которого ищется значение атрибута
    :param str attr: атрибут, значение которого необходимо получить (
        указывается полная цепочка, т.е. 'attr1.attr2.atr3')
    :param object default: значение по умолчанию
    :return: значение указанного атрибута или значение по умолчанию, если
        атрибут не был найден
    """
    try:
        value = operator.attrgetter(attr)(obj)
    except (
        AttributeError,
        ValueError,
        ObjectDoesNotExist,
    ):
        value = default

    return value


def date2str(date, template=None):
    """
    datetime.strftime глючит с годом < 1900
    типа обходной маневр (взято из django)

    WARNING from django:
    # This library does not support strftime's \"%s\" or \"%y\" format strings.
    # Allowed if there's an even number of \"%\"s because they are escaped.
    """
    return datetime_safe.new_datetime(date).strftime(
        template or
        settings.DATE_FORMAT or
        '%d.%m.%Y'
    )


def rebind_model_rel_id(obj):
    """
    Функция установки идентификатора объекта выступающего в роли внешней связи. Для FK-полей, если сохранили внешнюю
    модель, то проставим значение id в поле.

    Args:
        obj: объект, у которого должно быть произведено проставление идентификаторов объектов внешних связей
    """
    assert isinstance(obj, Model)

    for field in obj._meta.concrete_fields:
        if (
            field.is_relation
            and not getattr(obj, field.attname, None)
            and deep_getattr(obj, f'{field.name}.pk')
        ):
            # Получение объекта-значения внешнего ключа
            saved_obj = getattr(obj,  field.name)

            # Установка идентификатора объекта-значения внешнего ключа
            setattr(obj, field.attname, deep_getattr(obj, f'{field.name}.pk'))

            # Повторная установка объекта-значения внешнего ключа, т.к. при установке идентификатора, могла быть
            # произведена замена объекта. Такое поведение было замечено для моделей с выстроенной иерархией наследования
            setattr(obj, field.name, saved_obj)



from typing import (
    List,
    Union,
)

from function_tools.errors import (
    BaseError,
)


class BaseRunnableResult:
    """
    Базовый класс результатов выполнения запускаемых объектов
    """

    def __init__(self, *args, **kwargs):
        # Ключ результата, используемый для идентификации результата
        self._key = ''
        self._message = ''

        self._entities: List[Union[BaseError, 'BaseRunnableResult']] = []

    @property
    def entities(self) -> List[Union[BaseError, 'BaseRunnableResult']]:
        """
        Возвращает все сущности результата
        """
        return self._entities

    @property
    def errors(self) -> List[BaseError]:
        """
        Возвращает ошибки из результата. Т.к. в качестве сущностей могут
        выступать как ошибки, так и результаты выполнения запускаемых объектов,
        предусмотрен рекурсивный поиск ошибок.
        """
        errors_list = []

        for entity in self._entities:
            if isinstance(entity, BaseError):
                errors_list.append(entity)
            elif isinstance(entity, BaseRunnableResult):
                errors_list.extend(entity.errors)

        return errors_list

    @property
    def has_errors(self) -> bool:
        """
        Показывает, если ли в результате ошибки
        """
        return bool(self.errors)

    @property
    def has_not_errors(self) -> bool:
        """
        Показывает, нет ли в результате ошибок
        """
        return not bool(self.errors)

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, key):
        self._key = key

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, message_):
        self._message = message_

    def append_entity(self, entity: Union[BaseError, 'BaseRunnableResult']):
        """
        Добавление сущности в результат
        """
        if entity not in self._entities:
            self._entities.append(entity)



# Generated by Django 2.2.4 on 2021-09-05 23:05

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RegisteredFunction',
            fields=[
                ('function_path', models.CharField(db_index=True, max_length=1024, primary_key=True, serialize=False, unique=True, verbose_name='Путь до функции')),
                ('verbose_name', models.CharField(default='Имя функции не определено', max_length=512, verbose_name='Название функции')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Дата и время создания')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Дата и время обновления')),
            ],
            options={
                'verbose_name': 'Зарегистрированная функция',
                'db_table': 'function_tools_registered_function',
            },
        ),
    ]
